#!/usr/bin/env python3

import secube, getpass
import time
import sys

SE3_ALGO_AES = 0
SE3_ALGO_SHA256 = 1
SE3_ALGO_HMACSHA256 = 2
SE3_ALGO_AES_HMACSHA256 = 3
SE3_ALGO_MAX = 8

SE3_FEEDBACK_ECB = 1
SE3_FEEDBACK_CBC = 2
SE3_FEEDBACK_OFB = 3
SE3_FEEDBACK_CTR = 4
SE3_FEEDBACK_CFB = 5

SE3_DIR_SHIFT = 8
SE3_DIR_ENCRYPT = (1 << SE3_DIR_SHIFT)
SE3_DIR_DECRYPT = (2 << SE3_DIR_SHIFT)

SE3_CRYPTO_FLAG_FINIT = (1 << 15)
SE3_CRYPTO_FLAG_RESET = (1 << 14)
SE3_CRYPTO_FLAG_SETIV = SE3_CRYPTO_FLAG_RESET
SE3_CRYPTO_FLAG_SETNONCE = (1 << 13)
SE3_CRYPTO_FLAG_AUTH = (1 << 12)

def main():

    print("NOW %d"%int(time.time()))
    devs=secube.SEcube.discover()
    print("Found %d devices"%len(devs))
    for dev in devs:
        print(dev)

    if len(devs)>0:
        cube=secube.SEcube(devs[0])

        #password=getpass.getpass()
        password = "test"

        cube.login(password)

        string2Cypher = sys.argv[1]
        stringCyphered = None
        stringDecyphered = None
        print ()
        print ()
        print("[provaEncDec] String da cifrare -> " + string2Cypher)

        algorithm = SE3_ALGO_AES
        mode_enc = SE3_DIR_ENCRYPT | SE3_FEEDBACK_ECB
        flags_enc = mode_enc | SE3_CRYPTO_FLAG_FINIT
        mode_dec = SE3_DIR_DECRYPT | SE3_FEEDBACK_ECB
        flags_dec = mode_dec | SE3_CRYPTO_FLAG_FINIT
        key_id = 1

        lenIn = len(string2Cypher)
        lenOut = 0

        dataout=""

        try:
            print ()
            print ()
            cube.crypto_set_time(int(time.time()))
            print("[provaEncDec - setTime] Time setted")

            #encryption
            print ()
            print ()

            sess_id=cube.crypto_init(algorithm=algorithm, mode=mode_enc, key_id=key_id)
            print("[provaEncDec - encryption] crypto_init ok")

            print(bytearray(string2Cypher.encode()))

            stringCyphered = cube.crypto_update(sess_id=sess_id, flags=flags_enc, data1=None, data2=bytes(string2Cypher.encode()))
            print("[provaEncDec - encryption] crypto_update ok")
            print("[provaEncDec - encryption] Stringa cifrata: " + "".join(map(chr, stringCyphered)))

            #decryption
            print ()
            print ()

            sess_id=cube.crypto_init(algorithm=algorithm, mode=mode_dec, key_id=key_id)
            print("[provaEncDec - decryption] crypto_init ok")

            stringDecyphered = cube.crypto_update(sess_id=sess_id, flags=flags_dec, data1=None, data2=stringCyphered)
            print("[provaEncDec - decryption] crypto_update ok")
            print("[provaEncDec - decryption] Stringa decifrata: " + "".join(map(chr, stringDecyphered)))

            stringDecypheredJoin = "".join(map(chr, stringDecyphered))

            if(string2Cypher == stringDecypheredJoin):
                print("Matched")

            else:
                print("No Match")

        finally:
            cube.logout()
            cube.close()

        return 0

if __name__=="__main__":
    exit(main())
